import collections.abc
import numbers
import os
import random
import struct
import sys
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import flatbuffers
import numpy as np
import onnx
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from onnx import numpy_helper

# Generated by `flatc`
sys.path.append(os.path.dirname(__file__))
import FlatBuffers.Dl.AlignedBytes as AlignedBytes
import FlatBuffers.Dl.Attribute as Attribute
import FlatBuffers.Dl.AttributeF as AttributeF
import FlatBuffers.Dl.AttributeI as AttributeI
import FlatBuffers.Dl.AttributeType as AttributeType
import FlatBuffers.Dl.Dimension as Dimension
import FlatBuffers.Dl.DimensionValue as DimensionValue
import FlatBuffers.Dl.DimensionValueType as DimensionValueType
import FlatBuffers.Dl.Function as Function
import FlatBuffers.Dl.Graph as Graph
import FlatBuffers.Dl.Model as Model
import FlatBuffers.Dl.Node as Node
import FlatBuffers.Dl.OperatorSetId as OperatorSetId
import FlatBuffers.Dl.StringStringEntry as StringStringEntry
import FlatBuffers.Dl.Tensor as Tensor
import FlatBuffers.Dl.TensorDataType as TensorDataType
import FlatBuffers.Dl.TensorShape as TensorShape
import FlatBuffers.Dl.TensorTypeAndShape as TensorTypeAndShape
import FlatBuffers.Dl.TypeInfo as TypeInfo
import FlatBuffers.Dl.TypeInfoValue as TypeInfoValue
import FlatBuffers.Dl.ValueInfo as ValueInfo
import mapping as mapping

global_fbs_builder = None


def get_global_fbs_builder():
    global global_fbs_builder
    if global_fbs_builder is None:
        global_fbs_builder = flatbuffers.Builder(0)
    return global_fbs_builder


def reset_global_fbs_builder():
    global global_fbs_builder
    global_fbs_builder = None


def make_node(
    op_type: str,
    inputs: Sequence[str],
    outputs: Sequence[str],
    name: Optional[str] = None,
    doc_string: Optional[str] = None,
    domain: Optional[str] = None,
    **kwargs: Any,
) -> int:
    """Construct a Node.

    Args:
        op_type (string): The name of the operator to construct
        inputs (list of string): list of input names
        outputs (list of string): list of output names
        name (string, default None): optional unique identifier for NodeFbs
        doc_string (string, default None): optional documentation string for NodeFbs
        domain (string, default None): optional domain for NodeFbs.
            If it's None, we will just use default domain (which is empty)
        **kwargs (dict): the attributes of the node.  The acceptable values
            are documented in :func:`make_attribute`.

    Returns:
        Node fbs offset
    """
    builder = get_global_fbs_builder()
    op_type = builder.CreateString(op_type)

    inputs = [builder.CreateString(i) for i in inputs]
    Node.StartInputVector(builder, len(inputs))
    for i in reversed(inputs):
        builder.PrependUOffsetTRelative(i)
    inputs = builder.EndVector()

    outputs = [builder.CreateString(o) for o in outputs]
    Node.StartOutputVector(builder, len(outputs))
    for o in reversed(outputs):
        builder.PrependUOffsetTRelative(o)
    outputs = builder.EndVector()

    if name:
        name = builder.CreateString(name)
    if doc_string:
        doc_string = builder.CreateString(doc_string)
    if domain:
        domain = builder.CreateString(domain)

    attrs = None
    if kwargs:
        attrs = [make_attribute(key, value) for key, value in sorted(kwargs.items()) if value is not None]
        Node.StartAttributeVector(builder, len(attrs))
        for attr in reversed(attrs):
            builder.PrependUOffsetTRelative(attr)
        attrs = builder.EndVector()

    Node.Start(builder)
    Node.AddOpType(builder, op_type)
    Node.AddInput(builder, inputs)
    Node.AddOutput(builder, outputs)
    if name:
        Node.AddName(builder, name)
    if doc_string:
        Node.AddDocString(builder, doc_string)
    if domain:
        Node.AddDomain(builder, domain)
    if attrs is not None:
        Node.AddAttribute(builder, attrs)
    return Node.End(builder)


def make_graph_test_value(
    valuesForTest: Dict[str, Dict[str, np.ndarray]] = None, exponents: Dict[str, List[int]] = None
) -> Tuple[Sequence[int], Sequence[int]]:
    """Construct the test_inputs_value, test_outputs_value of Graph

    Args:
        valuesForTest (Dict[str, Dict[str, np.ndarray]]): the test values used to compare accuracy.
                                                          The input format is as follows:
            {
                'inputs': {
                    'input_0_name': np.ndarray
                    ......
                    'input_n_name': np.ndarray
                },
                'outputs': {
                    'output_0_name': np.ndarray
                    ......
                    'output_n_name': np.ndarray
                },
            }

    Returns:
        Tuple[Sequence[Tensor fbs offset], Sequence[Tensor fbs offset]]
    """
    input_test_values = []
    output_test_values = []
    if valuesForTest is not None:
        # Get the input test value
        for input_name, input_value in valuesForTest.get('inputs', {}).items():
            var_exponents = exponents.get(input_name, [0])
            tensor_type = np_dtype_to_tensor_dtype(input_value.dtype)
            input_shape = input_value.shape
            if len(input_shape) > 1:
                input_value = input_value.flatten()
            if len(input_shape) == 0:
                input_shape = [1]
            input_value_fbs = make_tensor(
                input_name, tensor_type, input_shape, input_value, raw=True, exponents=var_exponents
            )
            input_test_values.append(input_value_fbs)

        # Get the output test value
        for output_name, output_value in valuesForTest.get('outputs', {}).items():
            output_value = np.atleast_1d(output_value)
            var_exponents = exponents.get(output_name, [0])
            tensor_type = np_dtype_to_tensor_dtype(output_value.dtype)
            output_shape = output_value.shape
            if len(output_shape) > 1:
                output_value = output_value.flatten()
            output_value_fbs = make_tensor(
                output_name, tensor_type, output_shape, output_value, raw=True, exponents=var_exponents
            )
            output_test_values.append(output_value_fbs)

    return input_test_values, output_test_values


def make_graph(
    nodes: Sequence[int],
    name: str,
    inputs: Sequence[int],
    outputs: Sequence[int],
    initializer: Optional[Sequence[int]] = None,
    doc_string: Optional[str] = None,
    value_info: Optional[Sequence[int]] = None,
    test_inputs_value: Optional[Sequence[int]] = None,
    test_outputs_value: Optional[Sequence[int]] = None,
) -> int:
    """Construct a Graph

    Args:
        nodes: list of Node fbs offset
        name (string): graph name
        inputs: list of ValueInfo fbs offset
        outputs: list of ValueInfo fbs offset
        initializer: list of Tensor fbs offset
        doc_string (string): graph documentation
        value_info: list of ValueInfo fbs offset
    Returns:
        Graph fbs offset
    """
    builder = get_global_fbs_builder()

    name = builder.CreateString(name)
    if initializer is not None:
        Graph.StartInitializerVector(builder, len(initializer))
        for i in reversed(initializer):
            builder.PrependUOffsetTRelative(i)
        initializer = builder.EndVector()
    if value_info is not None:
        Graph.StartValueInfoVector(builder, len(value_info))
        for i in reversed(value_info):
            builder.PrependUOffsetTRelative(i)
        value_info = builder.EndVector()
    if test_inputs_value is not None:
        Graph.StartTestInputsValueVector(builder, len(test_inputs_value))
        for i in reversed(test_inputs_value):
            builder.PrependUOffsetTRelative(i)
        test_inputs_value = builder.EndVector()
    if test_outputs_value is not None:
        Graph.StartTestOutputsValueVector(builder, len(test_outputs_value))
        for i in reversed(test_outputs_value):
            builder.PrependUOffsetTRelative(i)
        test_outputs_value = builder.EndVector()

    Graph.StartNodeVector(builder, len(nodes))
    for i in reversed(nodes):
        builder.PrependUOffsetTRelative(i)
    nodes = builder.EndVector()

    Graph.StartInputVector(builder, len(inputs))
    for i in reversed(inputs):
        builder.PrependUOffsetTRelative(i)
    inputs = builder.EndVector()

    Graph.StartOutputVector(builder, len(outputs))
    for i in reversed(outputs):
        builder.PrependUOffsetTRelative(i)
    outputs = builder.EndVector()

    Graph.Start(builder)
    if initializer is not None:
        Graph.AddInitializer(builder, initializer)
    if value_info is not None:
        Graph.AddValueInfo(builder, value_info)
    if test_inputs_value is not None:
        Graph.AddTestInputsValue(builder, test_inputs_value)
    if test_outputs_value is not None:
        Graph.AddTestOutputsValue(builder, test_outputs_value)
    Graph.AddNode(builder, nodes)
    Graph.AddName(builder, name)
    Graph.AddInput(builder, inputs)
    Graph.AddOutput(builder, outputs)
    if doc_string:
        Graph.AddDocString(builder, builder.CreateString(doc_string))
    return Graph.End(builder)


def make_metadata_props(metadata_props: Dict[str, str]) -> Sequence[int]:
    """Construct a StringStringEntry

    Args:
        metadata_props (Dict[str, str]): the key/value of metadata_props
    Returns:
        StringStringEntry fbs offset
    """
    builder = get_global_fbs_builder()
    _metadata_props = []
    if metadata_props:
        for k, v in metadata_props.items():
            k_offset = builder.CreateString(k)
            v_offset = builder.CreateString(v)
            StringStringEntry.Start(builder)
            StringStringEntry.AddKey(builder, k_offset)
            StringStringEntry.AddValue(builder, v_offset)
            _metadata_props.append(StringStringEntry.End(builder))
    return _metadata_props


def make_model(graph: int, **kwargs: Any) -> bytes:
    """Construct a Model

    Args:
        graph (int): graph fbs offset
        **kwargs: any attribute to add to the returned instance
    Returns:
        model fbs data in bytes
    """
    builder = get_global_fbs_builder()

    opset_imports = kwargs.pop("opset_imports", None)  # type: ignore
    if opset_imports is not None:
        Model.StartOpsetImportVector(builder, len(opset_imports))
        for i in reversed(opset_imports):
            builder.PrependUOffsetTRelative(i)
        opset_imports = builder.EndVector()
    else:
        Model.StartOpsetImportVector(builder, 0)
        opset_imports = builder.EndVector()

    functions = kwargs.pop("functions", None)  # type: ignore
    if functions is not None:
        Model.StartFunctionsVector(builder, len(functions))
        for i in reversed(functions):
            builder.PrependUOffsetTRelative(i)
        functions = builder.EndVector()

    producer_name = kwargs.pop("producer_name", None)  # type: ignore
    if producer_name is not None:
        producer_name = builder.CreateString(producer_name)

    producer_version = kwargs.pop("producer_version", None)  # type: ignore
    if producer_version is not None:
        producer_version = builder.CreateString(producer_version)

    domain = kwargs.pop("domain", None)  # type: ignore
    if domain is not None:
        domain = builder.CreateString(domain)

    doc_string = kwargs.pop("doc_string", None)  # type: ignore
    if doc_string:
        doc_string = builder.CreateString(doc_string)

    metadata_props = kwargs.pop("metadata_props", None)
    if metadata_props:
        Model.StartMetadataPropsVector(builder, len(metadata_props))
        for i in reversed(metadata_props):
            builder.PrependUOffsetTRelative(i)
        metadata_props = builder.EndVector()

    Model.Start(builder)
    Model.AddGraph(builder, graph)
    Model.AddOpsetImport(builder, opset_imports)
    if functions is not None:
        Model.AddFunctions(builder, functions)
    ir_version = kwargs.pop("ir_version", None)  # type: ignore
    if ir_version is not None:
        Model.AddIrVersion(builder, ir_version)
    if producer_name:
        Model.AddProducerName(builder, producer_name)
    if producer_version:
        Model.AddProducerVersion(builder, producer_version)
    if domain:
        Model.AddDomain(builder, domain)
    model_version = kwargs.pop("model_version", None)  # type: ignore
    if model_version is not None:
        Model.AddModelVersion(builder, model_version)
    if doc_string:
        Model.AddDocString(builder, doc_string)
    if metadata_props:
        Model.AddMetadataProps(builder, metadata_props)
    model = Model.End(builder)
    builder.Finish(model)
    model_data = builder.Output()
    reset_global_fbs_builder()
    return model_data


def make_tensor(
    name: str,
    data_type: int,
    dims: Sequence[int],
    vals: Any,
    raw: bool = False,
    exponents: Sequence[int] = [0],
    doc_string: str = "",
) -> int:
    """Make a Tensor with specified arguments.  If raw is False, this
    function will choose the corresponding proto field to store the
    values based on data_type. If raw is True, use "raw_data" proto
    field to store the values, and values should be of type bytes in
    this case.

    Args:
        name (string): tensor name
        data_type (int): a value such as TensorDataType.FLOAT
        dims (List[int]): shape
        vals: values
        raw (bool): if True, vals contains the serialized content of the tensor,
            otherwise, vals should be a list of values of the type defined by *data_type*

    Returns:
        Tensor offset
    """
    builder = get_global_fbs_builder()
    name = builder.CreateString(name)
    # Check number of vals specified equals tensor size
    expected_size = 1
    if type(vals) is np.ndarray and len(vals.shape) > 1:
        vals = vals.flatten()
    for d in dims:
        expected_size *= d
    if len(vals) != expected_size:
        raise ValueError(
            f"Number of values does not match tensor's size. Expected {expected_size}, but it is {len(vals)}. "
        )
    dims = builder.CreateNumpyVector(np.array(dims, dtype=np.int64))
    exponents = builder.CreateNumpyVector(np.array(exponents, dtype=np.int64))
    if doc_string:
        doc_string = builder.CreateString(doc_string)

    if data_type == TensorDataType.TensorDataType().STRING and raw:
        raise TypeError("Can not use raw_data to store string type.")

    np_dtype = tensor_dtype_to_np_dtype(data_type)

    if raw:
        n = 16 // np_dtype.itemsize
        if len(vals) % n != 0:
            padding = n - len(vals) % n
            vals = np.pad(vals, (0, padding), mode='constant', constant_values=0)
        vals = np.reshape(vals, [-1, n])
        vals = vals.astype(np_dtype)
        Tensor.StartRawDataVector(builder, len(vals))
        for val in reversed(vals):
            AlignedBytes.CreateAlignedBytes(builder, val.tobytes())
        vals = builder.EndVector()
    else:
        if data_type == TensorDataType.TensorDataType().FLOAT16:
            raise TypeError("Don't support FLOAT16.")
        elif data_type == TensorDataType.TensorDataType().BOOL:
            vals = np.array(vals).astype(int)
        elif data_type == TensorDataType.TensorDataType().STRING:
            vals = np.array(vals).astype(bytes)
        else:
            vals = np.array(vals)
        vals = builder.CreateNumpyVector(vals)

    Tensor.Start(builder)
    Tensor.AddDataType(builder, data_type)
    Tensor.AddName(builder, name)
    if doc_string:
        Tensor.AddDocString(builder, doc_string)
    if raw:
        Tensor.AddRawData(builder, vals)
    else:
        add_field = tensor_dtype_to_field(data_type)
        add_field(builder, vals)
    Tensor.AddDims(builder, dims)
    Tensor.AddExponents(builder, exponents)
    return Tensor.End(builder)


def _to_bytes(value: Union[str, bytes]) -> bytes:
    """Coerce a string (or bytes) value into UTF-8 bytes."""
    return value if isinstance(value, bytes) else value.encode("utf-8")


def _to_str(value: Union[str, bytes]) -> str:
    """Coerce a string (or UTF-8 bytes) value into str."""
    return value if isinstance(value, str) else value.decode("utf-8", errors="ignore")


def make_attribute(
    key: str,
    value: Any,
    doc_string: Optional[str] = None,
    attr_type: Optional[int] = None,
) -> int:
    """Makes an Attribute based on the value type."""
    builder = get_global_fbs_builder()

    key = builder.CreateString(key)
    if doc_string:
        doc_string = builder.CreateString(doc_string)

    attr_i = None
    attr_f = None
    attr_s = None
    attr_t = None
    attr_ints = None
    attr_floats = None
    attr_strings = None
    # Singular cases
    if isinstance(value, numbers.Integral):
        attr_i = int(value)
    elif isinstance(value, numbers.Real):
        attr_f = float(value)
    elif isinstance(value, (str, bytes)):
        # Encode strings into utf-8
        value = _to_bytes(value)
        attr_s = builder.CreateByteVector(value)
    elif isinstance(value, onnx.TensorProto):
        value_data = numpy_helper.to_array(value)
        if (
            value.data_type >= TensorDataType.TensorDataType.UNDEFINED
            and value.data_type <= TensorDataType.TensorDataType.UINT64
        ):
            attr_t = make_tensor(
                name=value.name,
                data_type=value.data_type,
                dims=value.dims,
                vals=value_data,
                raw=True,
                doc_string=value.doc_string,
            )
        else:
            raise ValueError(f"Don't support type {onnx.TensorProto.DataType.Name(value.data_type)}")

    # Iterable cases
    elif isinstance(value, collections.abc.Iterable):
        value = list(value)
        if len(value) == 0 and attr_type is None:
            raise ValueError(f"Could not infer attribute `{key}` type from empty iterator")
        if attr_type is None:
            types = {type(v) for v in value}
            for exp_t, exp_enum in (
                (numbers.Integral, AttributeType.AttributeType().INTS),
                (numbers.Real, AttributeType.AttributeType().FLOATS),
                ((str, bytes), AttributeType.AttributeType().STRINGS),
            ):
                if all(issubclass(t, exp_t) for t in types):  # type: ignore[arg-type]
                    attr_type = exp_enum
                    break
            if attr_type is None:
                raise ValueError("Could not infer the attribute type from the elements of the passed Iterable value.")

        if attr_type == AttributeType.AttributeType().INTS:
            value = np.array(value, dtype=np.int64)
            attr_ints = builder.CreateNumpyVector(value)
        elif attr_type == AttributeType.AttributeType().FLOATS:
            value = np.array(value).astype(float)
            attr_floats = builder.CreateNumpyVector(value)
        elif attr_type == AttributeType.AttributeType().STRINGS:
            attr_strings = [builder.CreateString(s) for s in value]
            Attribute.StartStringsVector(builder, len(attr_strings))
            for s in reversed(attr_strings):
                builder.PrependUOffsetTRelative(s)
            attr_strings = builder.EndVector()
        else:
            raise AssertionError()  # Should not reach since `ValueError` must be raised in attr_type checking
    else:
        raise TypeError(f"'{value}' is not an accepted attribute value.")

    Attribute.AttributeStart(builder)
    Attribute.AddName(builder, key)
    if doc_string:
        Attribute.AddDocString(builder, doc_string)

    if attr_i is not None:
        Attribute.AddI(builder, AttributeI.CreateAttributeI(builder, attr_i))
        Attribute.AddAttrType(builder, AttributeType.AttributeType().INT)
    if attr_f is not None:
        Attribute.AddF(builder, AttributeF.CreateAttributeF(builder, attr_f))
        Attribute.AddAttrType(builder, AttributeType.AttributeType().FLOAT)
    if attr_s is not None:
        Attribute.AddS(builder, attr_s)
        Attribute.AddAttrType(builder, AttributeType.AttributeType().STRING)
    if attr_t is not None:
        Attribute.AddT(builder, attr_t)
        Attribute.AddAttrType(builder, AttributeType.AttributeType().TENSOR)
    if attr_ints is not None:
        Attribute.AddInts(builder, attr_ints)
        Attribute.AddAttrType(builder, AttributeType.AttributeType().INTS)
    if attr_floats is not None:
        Attribute.AddFloats(builder, attr_floats)
        Attribute.AddAttrType(builder, AttributeType.AttributeType().FLOATS)
    if attr_strings is not None:
        Attribute.AddStrings(builder, attr_strings)
        Attribute.AddAttrType(builder, AttributeType.AttributeType().STRINGS)
    return Attribute.End(builder)


def make_tensor_type_info(
    elem_type: int,
    shape: Optional[Sequence[Union[str, int, None]]],
    shape_denotation: Optional[List[str]] = None,
) -> int:
    builder = get_global_fbs_builder()
    tensor_shape = None
    if shape is not None:
        if shape_denotation and len(shape_denotation) != len(shape):
            raise ValueError("Invalid shape_denotation. Must be of the same length as shape.")
        dim_value = []
        for i, d in enumerate(shape):
            DimensionValue.Start(builder)
            if d is None:
                pass
            elif isinstance(d, int):
                DimensionValue.AddDimType(builder, DimensionValueType.DimensionValueType().VALUE)
                DimensionValue.AddDimValue(builder, d)

            elif isinstance(d, str):
                DimensionValue.AddDimType(builder, DimensionValueType.DimensionValueType().PARAM)
                DimensionValue.AddDimParam(builder, d)
            else:
                raise ValueError(f"Invalid item in shape: {d}. Needs to be of int or str.")
            dim_value.append(DimensionValue.End(builder))

        if shape_denotation is not None:
            shape_denotation = [builder.CreateString(builder, s) for s in shape_denotation]
        dim = []
        for i in range(len(shape)):
            Dimension.Start(builder)
            Dimension.AddValue(builder, dim_value[i])
            if shape_denotation is not None:
                Dimension.AddDenotation(builder, shape_denotation[i])
            dim.append(Dimension.End(builder))

        TensorShape.TensorShapeStartDimVector(builder, len(shape))
        for d in reversed(dim):
            builder.PrependUOffsetTRelative(d)
        dim = builder.EndVector()
        TensorShape.TensorShapeStart(builder)
        TensorShape.AddDim(builder, dim)
        tensor_shape = TensorShape.End(builder)

    TensorTypeAndShape.Start(builder)
    if tensor_shape is not None:
        TensorTypeAndShape.AddShape(builder, tensor_shape)
    TensorTypeAndShape.AddElemType(builder, elem_type)
    tensor_type_and_shape = TensorTypeAndShape.End(builder)

    TypeInfo.Start(builder)
    TypeInfo.AddValueType(builder, TypeInfoValue.TypeInfoValue().tensor_type)
    TypeInfo.AddValue(builder, tensor_type_and_shape)
    return TypeInfo.End(builder)


def make_tensor_value_info(
    name: str,
    elem_type: int,
    shape: Optional[Sequence[Union[str, int, None]]],
    doc_string: str = "",
    shape_denotation: Optional[List[str]] = None,
    exponents: Sequence[int] = [0],
) -> int:
    """Makes a ValueInfo fbs offset based on the data type and shape."""
    builder = get_global_fbs_builder()
    name = builder.CreateString(name)
    if doc_string:
        doc_string = builder.CreateString(doc_string)
    type_info = make_tensor_type_info(elem_type, shape, shape_denotation)
    exponents = builder.CreateNumpyVector(np.array(exponents, dtype=np.int64))
    ValueInfo.Start(builder)
    ValueInfo.AddName(builder, name)
    if doc_string:
        ValueInfo.AddDocString(builder, doc_string)
    ValueInfo.AddValueInfoType(builder, type_info)
    ValueInfo.AddExponents(builder, exponents)
    return ValueInfo.End(builder)


def aes_encryptor(data: bytes):
    key = bytearray(random.getrandbits(8) for _ in range(16))
    iv = bytearray(range(16, 32, 1))
    cipher = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
    encryptor = cipher.encryptor()
    ct_data = encryptor.update(data) + encryptor.finalize()

    hex_key = ''
    for _, byte in enumerate(key):
        hex_key += "0x{:02x}, ".format(byte)

    return ct_data, hex_key


def aes_decryptor(data: bytes, key: bytearray):
    iv = bytearray(range(16, 32, 1))
    cipher = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
    decryptor = cipher.decryptor()
    decrypted_data = decryptor.update(data) + decryptor.finalize()
    return decrypted_data


def save(data: bytes, file_path: str, encrypt_data: bool = False):
    key = None
    if encrypt_data:
        data, key = aes_encryptor(data)

    # file header 16 bytes
    header = bytearray("EDL2", 'utf-8')
    header += struct.pack("I", 1 if encrypt_data else 0)
    header += struct.pack("I", len(data))
    header += struct.pack("x") * 4
    with open(file_path, mode="wb") as f:
        f.write(header + data)
    return key


def _sanitize_str(s: Union[str, bytes]) -> str:
    if isinstance(s, str):
        sanitized = s
    elif isinstance(s, bytes):
        sanitized = s.decode("utf-8", errors="ignore")
    else:
        sanitized = str(s)
    if len(sanitized) < 64:  # noqa: PLR2004
        return sanitized
    return sanitized[:64] + f"...<+len={(len(sanitized) - 64)}>"


def printable_attribute(
    attr: Attribute.Attribute, subgraphs: bool = False
) -> Union[str, Tuple[str, List[Graph.Graph]]]:
    content = []
    content.append(_to_str(attr.Name()))
    content.append("=")

    def str_float(f: float) -> str:
        # NB: Different Python versions print different numbers of trailing
        # decimals, specifying this explicitly keeps it consistent for all
        # versions
        return f"{f:.15g}"

    def str_int(i: int) -> str:
        return str(i)

    _T = TypeVar("_T")

    def str_list(str_elem: Callable[[_T], str], xs: Sequence[_T]) -> str:
        return "[" + ", ".join(map(str_elem, xs)) + "]"

    # for now, this logic should continue to work as long as we are running on a flatbuffers
    # implementation. If/when we switch to flatbuffers, we will need to use attr.type

    # To support printing subgraphs, if we find a graph attribute, print out
    # its name here and pass the graph itself up to the caller for later
    # printing.

    attr_type = attr.AttrType()

    graphs = []
    if attr_type == AttributeType.AttributeType.INT:
        content.append(str_int(attr.I().I()))
    elif attr_type == AttributeType.AttributeType.FLOAT:
        content.append(str_float(attr.F().F()))
    elif attr_type == AttributeType.AttributeType.STRING:
        # TODO: Bit nervous about Python 2 / Python 3 determinism implications
        content.append(repr(_sanitize_str(attr.SAsNumpy().tobytes())))
    elif attr_type == AttributeType.AttributeType.TENSOR:
        if attr.T().DimsLength() > 0:
            content.append("<Tensor>")
        else:
            # special case to print scalars
            field = tensor_dtype_to_field(attr.t().DataType())
            content.append(f"<Scalar Tensor {getattr(attr.t(), field)}>")
    elif attr_type == AttributeType.AttributeType.FLOATS:
        content.append(str_list(str_float, [attr.Floats(i) for i in range(attr.FloatsLength())]))
    elif attr_type == AttributeType.AttributeType.INTS:
        content.append(str_list(str_int, [attr.Ints(i) for i in range(attr.IntsLength())]))
    elif attr_type == AttributeType.AttributeType.STRINGS:
        # TODO: Bit nervous about Python 2 / Python 3 determinism implications
        content.append(str(list(map(_sanitize_str, [attr.Strings(i) for i in range(attr.StringsLength())]))))
    elif attr_type == AttributeType.AttributeType.TENSORS:
        content.append("[<Tensor>, ...]")
    else:
        content.append("<Unknown>")
    if subgraphs:
        return " ".join(content), graphs
    return " ".join(content)


def printable_dim(dim: Dimension.Dimension) -> str:
    if dim.Value().DimType() == DimensionValueType.DimensionValueType.UNKNOWN:
        return "?"
    elif dim.Value().DimType() == DimensionValueType.DimensionValueType.VALUE:
        which = dim.Value().DimValue()
    elif dim.Value().DimType() == DimensionValueType.DimensionValueType.PARAM:
        which = dim.Value().DimParam()
    return str(which)


def get_class_attribute_name(obj: Any, value: int) -> str:
    ret: str = ""
    # 获取类和实例的所有属性名字
    all_attributes = dir(obj)
    filtered_attributes = [attr for attr in all_attributes if not attr.startswith('__') and not attr.endswith('__')]
    for attr in filtered_attributes:
        if getattr(obj, attr) == value:
            ret = attr
            break

    return ret


def printable_type(t: TypeInfo.TypeInfo) -> str:
    if t.ValueType() == TypeInfoValue.TypeInfoValue.tensor_type:
        tensor_type_and_shape = TensorTypeAndShape.TensorTypeAndShape()
        tensor_type_and_shape.Init(t.Value().Bytes, t.Value().Pos)
        s = get_class_attribute_name(TensorDataType.TensorDataType(), tensor_type_and_shape.ElemType())
        if tensor_type_and_shape.Shape() is not None:
            if tensor_type_and_shape.Shape().DimLength():
                s += str(
                    ", "
                    + "x".join(
                        map(
                            printable_dim,
                            [
                                tensor_type_and_shape.Shape().Dim(i)
                                for i in range(tensor_type_and_shape.Shape().DimLength())
                            ],
                        )
                    )
                )
            else:
                s += ", scalar"
        return s  # type: ignore[no-any-return]
    if t.ValueType() == TypeInfoValue.TypeInfoValue.NONE:
        return ""
    return f"Unknown type {t.valueType}"


def printable_value_info(v: ValueInfo.ValueInfo) -> str:
    s = f"%{_to_str(v.Name())}"
    if v.ValueInfoType():
        s = f"{s}[{printable_type(v.ValueInfoType())}]"
    if v.ExponentsLength() > 0:
        s += f", exponents: {repr(v.ExponentsAsNumpy().tolist())}"
    return s


def printable_tensor(t: Tensor.Tensor) -> str:
    s = f"%{_to_str(t.Name())}["
    s += get_class_attribute_name(TensorDataType.TensorDataType(), t.DataType())
    if t.DimsLength():
        s += str(", " + "x".join(map(str, [t.Dims(i) for i in range(t.DimsLength())])))
    else:
        s += ", scalar"
    s += "]"
    return s


def printable_tensor_value(t: Tensor.Tensor) -> str:
    s = f"%{_to_str(t.Name())}, "
    if not t.DimsIsNone():
        s += f"shape: {repr(t.DimsAsNumpy().tolist())}, "
    if not t.ExponentsIsNone():
        s += f"exponents: {repr(t.ExponentsAsNumpy().tolist())}, "
    if t.DocString() is not None:
        s += f"docString: {t.DocString()}, "

    if not t.FloatDataIsNone():
        s += f"value: {repr(t.FloatDataAsNumpy().tolist())}"
    elif not t.Int32DataIsNone():
        s += f"value: {repr(t.Int32DataAsNumpy().tolist())}"
    elif not t.StringDataIsNone():
        string_list = [t.StringData(i) for i in range(t.StringDataLength())]
        s += f"value: {repr(string_list)}"
    elif not t.Int64DataIsNone():
        s += f"value: {repr(t.Int64DataAsNumpy().tolist())}"
    elif not t.RawDataIsNone():
        np.set_printoptions(threshold=np.inf)
        chunks = []
        for i in range(t.RawDataLength()):
            chunk = t.RawData(i)
            chunks.append(np.frombuffer(chunk.BytesAsNumpy().tobytes(), dtype=tensor_dtype_to_np_dtype(t.DataType())))
        chunks = np.array(chunks).flatten()
        s += f"\nvalue: {repr(chunks)}"
    elif not t.DoubleDataIsNone():
        s += f"value: {repr(t.DoubleDataAsNumpy().tolist())}"
    elif not t.Uint64DataIsNone():
        s += f"value: {repr(t.Uint64DataAsNumpy().tolist())}"

    return s


def printable_node(
    node: Node.Node, prefix: str = "", subgraphs: bool = False
) -> Union[str, Tuple[str, List[Graph.Graph]]]:
    content = []
    if node.OutputLength():
        content.append(", ".join([f"%{_to_str(node.Output(i))}" for i in range(node.OutputLength())]))
        content.append("=")
    # To deal with nested graphs
    graphs: List[Graph.Graph] = []
    printed_attrs = []
    if node.AttributeLength():
        for i in range(node.AttributeLength()):
            attr = node.Attribute(i)
            if subgraphs:
                printed_attr_subgraphs = printable_attribute(attr, subgraphs)
                if not isinstance(printed_attr_subgraphs[1], list):
                    raise TypeError(f"printed_attr_subgraphs[1] must be an instance of {list}.")
                graphs.extend(printed_attr_subgraphs[1])
                printed_attrs.append(printed_attr_subgraphs[0])
            else:
                printed = printable_attribute(attr)
                if not isinstance(printed, str):
                    raise TypeError(f"printed must be an instance of {str}.")
                printed_attrs.append(printed)
    printed_attributes = ", ".join(sorted(printed_attrs))
    printed_inputs = ", ".join([f"%{_to_str(node.Input(i))}" for i in range(node.InputLength())])
    if node.AttributeLength():
        content.append(f"{_to_str(node.OpType())}[{printed_attributes}]({printed_inputs})")
    else:
        content.append(f"{_to_str(node.OpType())}({printed_inputs})")
    if subgraphs:
        return prefix + " ".join(content), graphs
    return prefix + " ".join(content)


def printable_model_properties(model: Model.Model) -> str:
    s = ""
    s += "metadata_props:\n"
    for i in range(model.MetadataPropsLength()):
        metadata_prop: StringStringEntry.StringStringEntry = model.MetadataProps(i)
        k = metadata_prop.Key()
        v = metadata_prop.Value()
        s += f"{_to_str(k)}: {_to_str(v)}\n"
    return s


def printable_graph(
    model: bytes,
    prefix: str = "",
    print_initializer_value: bool = False,
    print_value_info: bool = False,
    print_test_value: bool = False,
) -> str:
    """Display a Graph as a string.

    Args:
        graph (Graph.GraphT): the graph to display
        prefix (string): prefix of every line

    Returns:
        string
    """
    model = Model.Model.GetRootAs(model, 0)
    graph = model.Graph()

    content = []
    content.append("model properties:\n" + printable_model_properties(model) + "\n")

    indent = prefix + "  "
    # header
    header = ["graph", _to_str(graph.Name())]
    initializers = {graph.Initializer(i).Name() for i in range(graph.InitializerLength())}
    if graph.InputLength():
        header.append("(")
        in_strs = []  # required inputs
        in_with_init_strs = []  # optional inputs with initializer providing default value
        for i in range(graph.InputLength()):
            inp = graph.Input(i)
            if inp.Name() not in initializers:
                in_strs.append(printable_value_info(inp))
            else:
                in_with_init_strs.append(printable_value_info(inp))
        if in_strs:
            content.append(prefix + " ".join(header))
            header = []
            for line in in_strs:
                content.append(prefix + "  " + line)
        header.append(")")

        if in_with_init_strs:
            header.append("optional inputs with matching initializers (")
            content.append(prefix + " ".join(header))
            header = []
            for line in in_with_init_strs:
                content.append(prefix + "  " + line)
            header.append(")")

        # from IR 4 onwards an initializer is not required to have a matching graph input
        # so output the name, type and shape of those as well
        if len(in_with_init_strs) < len(initializers):
            graph_inputs = {graph.Input(i).Name() for i in range(graph.InputLength())}
            init_strs = [
                printable_tensor(graph.Initializer(i))
                for i in range(graph.InitializerLength())
                if graph.Initializer(i).Name() not in graph_inputs
            ]
            header.append("initializers (")
            content.append(prefix + " ".join(header))
            header = []
            for line in init_strs:
                content.append(prefix + "  " + line)
            header.append(")")

    header.append("{")
    content.append(prefix + " ".join(header))
    graphs: List[Graph.Graph] = []
    # body
    for i in range(graph.NodeLength()):
        node = graph.Node(i)
        contents_subgraphs = printable_node(node, indent, subgraphs=True)
        if not isinstance(contents_subgraphs[1], list):
            raise TypeError(f"contents_subgraphs[1] must be an instance of {list}.")
        content.append(contents_subgraphs[0])
        graphs.extend(contents_subgraphs[1])
    # tail
    tail = ["return"]
    if graph.OutputLength():
        tail.append(", ".join([f"%{_to_str(graph.Output(i).Name())}" for i in range(graph.OutputLength())]))
    content.append(indent + " ".join(tail))
    # closing bracket
    content.append(prefix + "}")
    for g in graphs:
        content.append("\n" + printable_graph(g))

    if print_initializer_value:
        initializers_value = [printable_tensor_value(graph.Initializer(i)) for i in range(graph.InitializerLength())]
        content.append("\n\n" + "initializers value:\n" + "\n".join(initializers_value))

    if print_value_info:
        values_info = [printable_value_info(graph.ValueInfo(i)) for i in range(graph.ValueInfoLength())]
        content.append("\n\n" + "values info:\n" + "\n".join(values_info))

    if print_test_value:
        test_inputs_value = [
            printable_tensor_value(graph.TestInputsValue(i)) for i in range(graph.TestInputsValueLength())
        ]
        content.append("\n\n" + "test inputs value:\n" + "\n".join(test_inputs_value))
        test_outputs_value = [
            printable_tensor_value(graph.TestOutputsValue(i)) for i in range(graph.TestOutputsValueLength())
        ]
        content.append("\n\n" + "test outputs value:\n" + "\n".join(test_outputs_value))

    return "\n".join(content)


# Following functions are used for mapping
def tensor_dtype_to_np_dtype(tensor_dtype: int) -> np.dtype:
    """Convert a TensorDataType to corresponding numpy dtype. It can be used while making tensor.

    Args:
        tensor_dtype: TensorDataType

    Returns:
        numpy's data_type
    """
    return mapping.TENSOR_TYPE_MAP[tensor_dtype].np_dtype


def tensor_dtype_to_field(tensor_dtype: int) -> Callable[..., int]:
    """Convert a TensorDataType to corresponding field name for storage. It can be used while making tensors.

    Args:
        tensor_dtype: TensorDataType

    Returns:
        field name
    """
    return mapping._STORAGE_TENSOR_TYPE_TO_FIELD[mapping.TENSOR_TYPE_MAP[tensor_dtype].storage_dtype]


def np_dtype_to_tensor_dtype(np_dtype: np.dtype) -> int:
    """Convert a numpy's dtype to corresponding TensorDataType. It can be used while converting numpy arrays to tensors.

    Args:
        np_dtype: numpy's data_type

    Returns:
        TensorDataType
    """
    tensor_type = mapping._NP_TYPE_TO_TENSOR_TYPE.get(np_dtype)
    if tensor_type == None:
        return cast(int, TensorDataType.TensorDataType().UNDEFINED)
    else:
        return cast(int, tensor_type)
    # return cast(
    #     int,
    #     mapping._NP_TYPE_TO_TENSOR_TYPE[np_dtype],
    # )



import random
from typing import List

import numpy as np
import torch

from esp_ppq.core import TensorQuantizationConfig
from esp_ppq.executor import TorchQuantizeDelegator
from esp_ppq.quantization.qfunction import PPQLinearQuantFunction
from esp_ppq.utils.ema import EMARecorder


class BanditDelegator(TorchQuantizeDelegator):
    """带有多臂赌博机的量化代理，从 ppq 0.6.2 版本后，我们引入 多臂赌博机算法训练 scale 与 offset。在未来我们可能还会引入其他
    类似的算法，例如UCB，马尔可夫蒙特卡洛估计等。

    引入这些算法的原因是我们注意到 scale 与 offset 的导数非常不靠谱
    为此我们引入简单的强化学习，直接估计P(r | scale=s, context)
    即再给定上下文 context 的情况下，选取当前 scale 为 s，获利的概率

    Quantization with multi-arm bandit.

    Multi-arm bandits are introduced since PPQ 0.6.2 for training
        quantization scale and offset.
    """

    def __init__(self, arms: List[float], config: TensorQuantizationConfig) -> None:
        if len(arms) < 2:
            raise ValueError('Can not initialize bandit with less than 2 arms.')
        self.e = 0.1
        self.arms = arms
        self.num_of_arms = len(arms)
        self.rewards = [EMARecorder() for _ in range(self.num_of_arms)]
        self.rewards[0].push(1)
        self.last_selected = 0
        self.reference = config.scale.clone()
        self.config = config
        self.decay = 0.99

    def roll(self) -> int:
        if random.random() > self.e:
            selected = random.randint(0, len(self.arms) - 1)
        else:
            selected = np.argmax([ema.pop() for ema in self.rewards])
        self.last_selected = selected
        return selected

    def mark(self, rewards: float):
        self.rewards[self.last_selected].push(rewards)

    def finalize(self) -> bool:
        self.config.scale = self.reference * self.arms[np.argmax([ema.pop() for ema in self.rewards])]

    def withdraw(self):
        self.config.scale = self.reference

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        config.scale = self.reference * self.arms[self.roll()]
        return PPQLinearQuantFunction(tensor, config)



# 这里我们展示两种不同的方法去生成 TensorRT Engine

# Plan A: TensorRT 自己校准
import glob
import json
import os

import numpy as np
import onnx
import pycuda.autoinit  # noqa F401
import pycuda.driver as cuda
import tensorrt as trt
import torch
import torchvision.transforms as transforms
from PIL import Image

ONNX_PATH = 'models/yolov5s.v5.onnx'  # 你的模型位置
ENGINE_PATH = 'Output/yolov5s.v5(trt).engine'  # 生成的 Engine 位置
CALIBRATION_PATH = 'imgs'  # 校准数据集


class EntropyCalibrator(trt.IInt8MinMaxCalibrator):
    def __init__(self, files_path=r'imgs'):
        trt.IInt8MinMaxCalibrator.__init__(self)

        self.cache_file = 'Net.cache'

        self.batch_size = 1
        self.Channel = 3
        self.Height = 640
        self.Width = 640
        self.transform = transforms.Compose(
            [
                transforms.Resize([self.Height, self.Width]),  # [h,w]
                transforms.ToTensor(),
            ]
        )

        self.imgs = glob.glob(os.path.join(files_path, '*'))
        self.batch_idx = 0
        self.max_batch_idx = len(self.imgs) // self.batch_size
        self.data_size = trt.volume([self.batch_size, self.Channel, self.Height, self.Width]) * trt.float32.itemsize
        self.device_input = cuda.mem_alloc(self.data_size)

    def next_batch(self):
        if self.batch_idx < self.max_batch_idx:
            batch_files = self.imgs[self.batch_idx * self.batch_size : (self.batch_idx + 1) * self.batch_size]
            batch_imgs = np.zeros((self.batch_size, self.Channel, self.Height, self.Width), dtype=np.float32)
            for i, f in enumerate(batch_files):
                img = Image.open(f).convert('RGB')
                img = self.transform(img).numpy()
                assert img.nbytes == self.data_size / self.batch_size, 'not valid img!' + f
                batch_imgs[i] = img
            self.batch_idx += 1
            print("batch:[{}/{}]".format(self.batch_idx, self.max_batch_idx))
            return np.ascontiguousarray(batch_imgs)
        else:
            return np.array([])

    def get_batch_size(self):
        return self.batch_size

    def get_batch(self, names, p_str=None):
        try:
            batch_imgs = self.next_batch()
            if batch_imgs.size == 0 or batch_imgs.size != self.batch_size * self.Channel * self.Height * self.Width:
                return None
            cuda.memcpy_htod(self.device_input, batch_imgs.astype(np.float32))
            return [int(self.device_input)]
        except:
            return None

    def read_calibration_cache(self):
        # If there is a cache, use it instead of calibrating again. Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                return f.read()

    def write_calibration_cache(self, cache):
        with open(self.cache_file, "wb") as f:
            f.write(cache)


# create builder and network
device = torch.device('cuda')
logger = trt.Logger(trt.Logger.ERROR)
builder = trt.Builder(logger)
EXPLICIT_BATCH = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
network = builder.create_network(EXPLICIT_BATCH)
parser = trt.OnnxParser(network, logger)

onnx_model = onnx.load(ONNX_PATH)
parser.parse(onnx_model.SerializeToString())

config = builder.create_builder_config()
# config.max_workspace_size = 4 << 30 no need for trt 8.4GA
config.set_flag(trt.BuilderFlag.INT8)
calibrator = EntropyCalibrator(files_path=CALIBRATION_PATH)
config.int8_calibrator = calibrator

plan = builder.build_serialized_network(network, config)

with open(ENGINE_PATH, mode='wb') as f:
    f.write(plan)



import copy
import json
from typing import Union

from esp_ppq.core import *


# ----------------------------------
# Legacy Settings, Following Settings are not used in PPQ 0.6.5
class AdvancedOptimizationSetting:
    def __init__(self) -> None:
        self.collecting_device = 'executor'
        self.auto_check = False
        self.limit = 4.0
        self.lr = 1e-3
        self.steps = 2500
        self.interested_layers = []
        self.interested_outputs = []
        self.verbose = True


class MatrixFactorizationSetting:
    def __init__(self) -> None:
        self.interested_layers = []
        self.method = 'svd'


# ----------------------------------


class BlockwiseReconstructionSetting:
    def __init__(self) -> None:
        # 训练那些层，不设置则训练全部层
        self.interested_layers = []

        # scale 是否可以训练
        self.is_scale_trainable = False

        # 学习率
        self.lr = 1e-3

        # 学习步数
        self.steps = 5000

        # 正则化参数
        self.gamma = 1.0

        # 缓存设备
        self.collecting_device = 'cuda'

        # 区块大小
        self.block_size = 4


class SSDEqualizationSetting:
    def __init__(self) -> None:
        # Equalization 优化级别，目前只支持level 1，对应 Conv--Relu--Conv 和 Conv--Conv 的拉平
        # optimization level, only support level 1 for now
        # you shouldn't modify this
        self.opt_level = 1

        # 在计算scale的时候，所有低于 channel_ratio * max(W) 的值会被裁减到 channel_ratio * max(W)
        # channel ratio used to calculate equalization scale
        # all values below this ratio of the maximum value of corresponding weight
        # will be clipped to this ratio when calculating scale
        self.channel_ratio = 0.5

        # loss的降低阈值，优化后的loss低于 原来的loss * 降低阈值, 优化才会发生
        # optimized loss must be below this threshold of original loss for algo to take effect
        self.loss_threshold = 0.8

        # 是否对权重进行正则化
        # whether to apply layer normalization to weights
        self.layer_norm = False

        # 算法迭代次数，3次对于大部分网络足够
        # num of iterations, 3 would be enough for most networks
        # it takes about 10 mins for one iteration
        self.iteration = 3


class ChannelSplitSetting:
    def __init__(self) -> None:
        # channel split 优化级别，如果选成 1 则不进行多分支模式匹配，如果选成 2，则进行跨越 add, sub 的多分支模式匹配
        # 不一定哪一个好，你自己试试
        # optimization level of layerwise channel split
        # 1 - single branch channel split(can not cross add, sub)
        # 2 - multi branch channel split(channel split cross add, sub)
        # don't know which one is better, try it by yourself.
        self.opt_level = 1

        # channel split 迭代次数，试试 1，2，3，10，100
        # algorithm iteration times, try 1, 2, 3, 10, 100
        self.iterations = 10

        # channel split 权重阈值，试试 0.5, 2
        # 这是个十分重要的属性，所有小于该值的通道不会参与运算
        # value threshold of channel split, try 0.5 and 2
        # it is a curical setting of channel split, value below this threshold won't get included in this optimizition.
        self.value_threshold = 0.5  # try 0.5 and 2, it matters.

        # 是否在 channel split 中考虑 bias
        # whether to equalize bias as well as weight
        self.including_bias = False
        self.bias_multiplier = 0.5

        # 是否在 channel split 中考虑 activation
        # whether to equalize activation as well as weight
        self.including_act = False
        self.act_multiplier = 0.5

        # Only layer that listed in interested_layers will be processed by this pass.
        # If interested_layers is None or empty list, all the layers will be processed.
        self.interested_layers = []


class BiasCorrectionSetting:
    def __init__(self) -> None:
        # 指定所有需要执行 BiasCorrection 的层的名字，不写就是所有层全部进行 bias correction
        self.interested_layers = []

        # 指定 BiasCorrection 的区块大小，越大越快，但也越不精准
        self.block_size = 4

        # 指定 bias 的统计步数，越大越慢，越小越不精准
        self.steps = 32

        # 缓存数据放在哪
        self.collecting_device = 'executor'


class WeightSplitSetting:
    def __init__(self) -> None:
        # 所有需要分裂的层的名字，Weight Split 会降低网络执行的性能，你必须手动指定那些层要被分裂
        # Weight Split 和 Channel Split 都是一种以计算时间作为代价，提高量化精度的方法
        # 这些方法主要适用于 per-tensor 的量化方案
        # computing layers which are intended to be splited
        self.interested_layers = []

        # 所有小于阈值的权重将被分裂
        self.value_threshold = 2.0

        # 分裂方式，可以选 balance(平均分裂), random(随机分裂)
        self.method = 'balance'


class GraphFormatSetting:
    def __init__(self) -> None:
        # 有一些平台不支持Constant Op，这个pass会尝试将 Constant Operation 的输入转变为 Parameter Variable
        # Some deploy platform does not support constant operation,
        # this pass will convert constant operation into parameter variable(if can be done).
        self.format_constant_op = True

        # 融合Conv和Batchnorm
        # Fuse Conv and Batchnorm Layer. This pass is necessary and crucial.
        self.fuse_conv_bn = True

        self.fuse_swish = True

        # 将所有的parameter variable进行分裂，使得每个variable具有至多一个输出算子
        # Split all parameter variables, making all variable has at most 1 output operation.
        # This pass is necessary and crucial.
        self.format_paramters = True

        # 一个你必须要启动的 Pass
        # This pass is necessary and crucial.
        self.format_cast = True

        # 尝试从网络中删除所有与输出无关的算子和 variable
        # Remove all unnecessary operations and variables(which are not link to graph output) from current graph,
        # notice that some platform use unlinked variable as quantization parameters, do not set this as true if so.
        self.delete_isolate = True


class EqualizationSetting:
    def __init__(self) -> None:
        # Equalization 优化级别，如果选成 1 则不进行多分支拉平，如果选成 2，则进行跨越 add, sub 的多分支拉平
        # 不一定哪一个好，你自己试试
        # optimization level of layerwise equalization
        # 1 - single branch equalization(can not cross add, sub)
        # 2 - multi branch equalization(equalization cross add, sub)
        # don't know which one is better, try it by yourself.
        self.opt_level = 1

        # Equalization 迭代次数，试试 1，2，3，10，100
        # algorithm iteration times, try 1, 2, 3, 10, 100
        self.iterations = 10

        # Equalization 权重阈值，试试 0.5, 2
        # 这是个十分重要的属性，所有小于该值的权重不会参与运算
        # value threshold of equalization, try 0.5 and 2
        # it is a curical setting of equalization, value below this threshold won't get included in this optimizition.
        self.value_threshold = 0.5  # try 0.5 and 2, it matters.

        # 是否在 Equalization 中拉平 bias
        # whether to equalize bias as well as weight
        self.including_bias = False
        self.bias_multiplier = 0.5

        # 是否在 Equalization 中拉平 activation
        # whether to equalize activation as well as weight
        self.including_act = False
        self.act_multiplier = 0.5

        # Only layer that listed in interested_layers will be processed by this pass.
        # If interested_layers is None or empty list, all the layers will be processed.
        self.interested_layers = []


class ActivationQuantizationSetting:
    def __init__(self) -> None:
        # 激活值校准算法，不区分大小写，可以选择 minmax, kl, percentile, MSE, None
        # 选择 None 时，将由 quantizer 指定量化算法
        # activation calibration method
        self.calib_algorithm = None


class QuantConfigModifySetting:
    def __init__(self) -> None:
        self.custom_config = {}


class ParameterQuantizationSetting:
    def __init__(self) -> None:
        # 参数校准算法，不区分大小写，可以选择 minmax, percentile, kl, MSE, None
        # parameter calibration method
        self.calib_algorithm = None

        # 是否处理被动量化参数
        # whether to process passive parameters
        self.quantize_passive_parameter = True
        # Whether to export quant info of clip min, max
        self.clip_visiblity: QuantizationVisibility = QuantizationVisibility.INTERNAL
        # Whether to export quant info of pad value
        self.pad_visiblity: QuantizationVisibility = QuantizationVisibility.INTERNAL

        # 是否执行参数烘焙
        # whether to bake quantization on parameter.
        self.baking_parameter = True


class QuantizationFusionSetting:
    def __init__(self) -> None:
        # 一个你必须要启动的 Pass，删除无效定点。
        # PPQ 的定点过程与硬件设备一致，每一个算子都会尝试对它的输入和输出进行定点
        # 但在网络当中，当输入已经被上游算子进行了定点，则当前算子无需再对其输入进行定点
        # This pass is necessary and curicial, remove all unnecessary quantization from your graph.
        # PPQ executor quantizes each input tensor and output tensor of current operation.
        # however if input tensor has been quantized by upstream operations, then there is no need to quantize it again.
        self.remove_useless_quantization = True

        # computing operation - activation operation 联合定点
        # joint quantization with computing operations and following activations.
        self.fuse_activation = True

        # 省略被动算子的定点，保持算子前后定点信息一致
        # skip passive operation's input and output quantization, keep them with a same quantization scale and offset.
        self.fuse_passive_op = True

        # 对多输入算子执行定点对齐
        # 对于 Add, Concat 这类具有多个输入的算子，硬件执行时必须要求所有输入位宽相同且量化信息一致
        # 在 8 bit 网络中，我们不讨论位宽的问题，只讨论量化信息
        # 选择对齐方式 - Align to Large, 则所有输入量化信息将合并成一个更大范围的量化信息,
        #   保证其能表示所有输入的最小值与所有输入的最大值。
        # 选择对齐方式 - Align to Output, 则所有输入量化信息将全部等同于算子输出的量化信息
        # 选择对齐方式 - None, 则不处理对应算子的量化对齐
        # 设置 force_alignment_overlap 为 True, 则强制覆盖算子的上游量化信息，对于特殊结构的网络，这可能导致大范围的合并。

        # For Operations like Add, Concat, hardware requires their inputs share a same quantization config.
        # So that we implements 2 alignment method here for simulating hardware behaviour:
        # Align to Large: all input variables will merge their quantization range to a larger one.
        # Align to Output: all input quantization config will be replaced by output quantization config.
        # None: take no alignment here, input and output will have their own(independent) quantization config.
        # force_alignment_overlap: if set to true, ppq will overlap upstream quantization config, for some networks
        #   it will incurs configuration coalesce among the whole graph.
        self.align_quantization = True
        self.align_avgpooling_to = 'None'
        self.align_elementwise_to = 'Align to Large'
        self.align_concat_to = 'Align to Output'
        self.align_resize_to = 'Align to Output'
        self.force_alignment_overlap = True


class LSQSetting:
    def __init__(self) -> None:
        # should only contain computing ops, if given, ops in interested_layers will be checked and
        # if conditions are satisfied, weight, scale of weight, scale of activation will be trained
        # with mse optimization goal, if not given, every condition-satisfied computing op will be
        # optimized
        self.interested_layers = []

        # initial learning rate, by default Adam optimizer and a multistep scheduler with 0.1 decay
        # are used for convergence
        self.lr = 1e-5

        # collecting device for block input and output
        # turn this to cpu if CUDA OOM Error
        self.collecting_device = 'cuda'

        # num of training steps, please adjust it to your needs
        self.steps = 500

        # is scale trainable
        self.is_scale_trainable = True

        # regularization term
        self.gamma = 0.0

        # block size of graph cutting.
        self.block_size = 4


class TemplateSetting:
    """TemplateSetting 只是为了向你展示如何创建一个新的 Setting 并传递给相对应的 pass 传递的过程在
    esp_ppq.quantization.quantizer.base.py 里面.

    TemplateSetting just shows you how to create a setting class.
        parameter passing is inside esp_ppq.quantization.quantizer.base.py

    Raises:
        TypeError: [description]

    Returns:
        [type]: [description]
    """

    def __init__(self) -> None:
        self.my_first_parameter = 'This Parameter just shows you how to create your own parameter.'


class DispatchingTable:
    def __init__(self) -> None:
        self.intro_0 = 'Dispatching Table is a str -> TargetPlatform dictionary.'
        self.intro_1 = 'Any thing listed in this table will override the policy of PPQ dispatcher.'
        self.intro_2 = 'Example above shows you how to edit a valid dispatching table.'
        self.attention = 'All operation names that can not be found with your graph will be ignored.'

        self.dispatchings = {
            'YOUR OEPRATION NAME': 'TARGET PLATFORM(INT)',
            'FP32 OPERATION NAME': TargetPlatform.FP32.value,
            'SOI OPERATION NAME': TargetPlatform.SOI.value,
            'DSP INT8 OPERATION NAME': TargetPlatform.PPL_DSP_INT8.value,
            'TRT INT8 OPERATION NAME': TargetPlatform.TRT_INT8.value,
            'NXP INT8 OPERATION NAME': TargetPlatform.NXP_INT8.value,
            'PPL INT8 OPERATION NAME': TargetPlatform.PPL_CUDA_INT8.value,
        }

    def append(self, operation: str, platform: Union[int, TargetPlatform]):
        assert isinstance(platform, int) or isinstance(platform, TargetPlatform), (
            'Your dispatching table contains a invalid setting of operation {operation}, '
            'All platform setting given in dispatching table is expected given as int or TargetPlatform instance, '
            f'however {type(platform)} was given.'
        )
        if isinstance(platform, TargetPlatform):
            platform = platform.value
        self.dispatchings.update({operation: platform})


class QuantizationSetting:
    def __init__(self) -> None:
        # 子图切分与调度算法，可从 'pointwise', 'conservative', 'pursus', 'allin', 'pplnn' 中五选一，不区分大小写
        self.dispatcher = 'conservative'

        self.graph_format_setting = GraphFormatSetting()

        # ssd with loss check equalization 相关设置
        # may take longer time(about 30min for default 3 iterations), but guarantee better result than baseline
        # should not be followed by a plain equalization when turned on
        self.ssd_equalization = False
        self.ssd_setting = SSDEqualizationSetting()

        # layer wise equalizition 与相关配置
        self.equalization = False
        self.equalization_setting = EqualizationSetting()

        self.weight_split = False
        self.weight_split_setting = WeightSplitSetting()

        # OCS channel split configuration
        self.channel_split = False
        self.channel_split_setting = ChannelSplitSetting()

        # Matrix Factorization Split. (Experimental method)
        self.matrix_factorization = False
        self.matrix_factorization_setting = MatrixFactorizationSetting()

        # activation 量化与相关配置
        self.quantize_activation = True
        self.quantize_activation_setting = ActivationQuantizationSetting()

        self.quant_config_modify = False
        self.quant_config_modify_setting = QuantConfigModifySetting()

        # 参数量化与相关配置
        self.quantize_parameter = True
        self.quantize_parameter_setting = ParameterQuantizationSetting()

        # 是否执行网络微调
        self.lsq_optimization = False
        self.lsq_optimization_setting = LSQSetting()

        self.blockwise_reconstruction = False
        self.blockwise_reconstruction_setting = BlockwiseReconstructionSetting()

        # 是否启动 bias correction pass
        self.bias_correct = False
        self.bias_correct_setting = BiasCorrectionSetting()

        # 量化融合相关配置
        self.fusion = True
        self.fusion_setting = QuantizationFusionSetting()

        # extension setting 只是一个空白的占位符，用来向你展示如何创建一个自定义的 setting 并传递参数。
        # extension setting shows you how to create a setting and pass parameter to passes.
        # see esp_ppq.quantization.quantizer.base.py
        self.extension = False
        self.extension_setting = TemplateSetting()

        # ConvTranspose 分解相关配置
        # ConvTranspose decomposition configuration
        self.convtranspose_decomposition = True

        # 程序签名
        self.version = PPQ_CONFIG.VERSION
        self.signature = PPQ_CONFIG.NAME

        # 算子调度表，你可以编辑它来手动调度算子。
        self.dispatching_table = DispatchingTable()

    def to_json(self) -> str:
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4, ensure_ascii=False)


class QuantizationSettingFactory:
    @staticmethod
    def default_setting() -> QuantizationSetting:
        return QuantizationSetting()

    @staticmethod
    def academic_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.fusion = False
        return default_setting

    @staticmethod
    def ncnn_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.fusion = False
        return default_setting

    @staticmethod
    def mnn_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.fusion = True
        return default_setting

    @staticmethod
    def pplcuda_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.equalization = False
        return default_setting

    @staticmethod
    def metax_pertensor_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.equalization = True
        return default_setting

    @staticmethod
    def ascend_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.fusion = False
        return default_setting

    @staticmethod
    def dsp_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.equalization = True
        default_setting.equalization_setting.opt_level = 1
        default_setting.equalization_setting.value_threshold = 0.5
        default_setting.equalization_setting.iterations = 3

        return default_setting

    @staticmethod
    def nxp_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.equalization = False
        return default_setting

    @staticmethod
    def fpga_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.equalization = True
        default_setting.equalization_setting.including_bias = True
        return default_setting

    @staticmethod
    def trt_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        return default_setting

    @staticmethod
    def espdl_setting() -> QuantizationSetting:
        default_setting = QuantizationSetting()
        default_setting.quantize_parameter_setting.clip_visiblity = QuantizationVisibility.EXPORT_WHEN_ACTIVE
        default_setting.quantize_parameter_setting.pad_visiblity = QuantizationVisibility.EXPORT_WHEN_ACTIVE
        default_setting.quantize_parameter_setting.baking_parameter = False
        default_setting.quantize_activation_setting.calib_algorithm = 'kl'
        default_setting.fusion_setting.align_elementwise_to = 'Align to Output'
        default_setting.fusion_setting.force_alignment_overlap = False
        default_setting.quant_config_modify = False
        return default_setting

    @staticmethod
    def from_json(json_obj: Union[str, dict]) -> QuantizationSetting:
        setting = QuantizationSetting()
        if isinstance(json_obj, str):
            setting_dict = json.loads(json_obj)
        else:
            setting_dict = json_obj

        if 'version' not in setting_dict:
            ppq_warning('Can not find a valid version from your json input, input might not be correctly parsed.')
        else:
            version = setting_dict['version']
            if version < PPQ_CONFIG.VERSION:
                ppq_warning(
                    f'You are loading a json quantization setting created by PPQ of another version: '
                    f'({version}), input setting might not be correctly parsed. '
                    'And all missing attributes will set as default without warning.'
                )

        if not isinstance(setting_dict, dict):
            raise TypeError(
                'Can not load setting from json string, '
                f'expect a dictionary deserialzed from json here, '
                f'however {type(setting_dict)} was given.'
            )

        def assign(obj_setting: dict, obj: object):
            for key, value in obj_setting.items():
                if key in obj.__dict__:
                    if 'builtin' in obj.__dict__[key].__class__.__module__:
                        obj.__dict__[key] = copy.deepcopy(value)
                    else:
                        assert isinstance(value, dict)
                        assign(value, obj.__dict__[key])
                else:
                    ppq_warning(
                        f'Unexpected attribute ({key}) was found in {obj.__class__.__name__}. '
                        'This might because you are using a setting generated by PPQ with another version. '
                        'We will continue setting loading progress, while this attribute will be skipped.'
                    )

        assign(setting_dict, setting)
        return setting



from typing import Any, List

import torch

from esp_ppq.core import TargetPlatform, convert_any_to_numpy, convert_any_to_torch_tensor

from .base.command import GraphCommand, GraphCommandType, GraphDeployCommand
from .base.graph import BaseGraph, Operation, Variable
from .processer import GraphCommandProcessor
from .quantize import QuantableOperation


class RunnableGraph(GraphCommandProcessor):
    def __init__(self, graph: BaseGraph, device: str = None):
        """RunnableGraph deals with values related with graph executing.

        Literally it helps you move values of your graph towards device and vice versa.
            And give an executable order of all operations in your graph which actual executor will follow.
        Args:
            graph (BaseGraph): BaseGraph instance.
            device (str, optional): This attribute is only used by with RunnableGraph(graph, device) syntactic.
            next_command_processor (Callable, optional): next processor in processing chain.
        """
        super().__init__(graph_or_processor=graph)
        self._device = device  # only used in "with RunnableGraph(graph, device):"

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.DEPLOY_TO_CPU:
            return self.deploy('cpu')

        elif command.command_type == GraphCommandType.DEPLOY_TO_CUDA:
            if isinstance(command, GraphDeployCommand):
                device = command._device
                return self.deploy(device)
            else:
                return self.deploy('cuda')

        elif command.command_type == GraphCommandType.DEPLOY_TO_NUMPY:
            return self.retrieve()

        elif command.command_type == GraphCommandType.DEPLOY_TO_MPS:
            return self.deploy('mps')

    def __enter__(self):
        self.deploy(self._device)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.retrieve()

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.DEPLOY_TO_CPU,
            GraphCommandType.DEPLOY_TO_CUDA,
            GraphCommandType.DEPLOY_TO_NUMPY,
            GraphCommandType.DEPLOY_TO_MPS,
        ]

    def retrieve(self):
        for _, operator in self._graph.operations.items():
            assert isinstance(operator, Operation), (
                f'Failed to retrieve graph to numpy, incorrect operator {operator} found.'
            )

            # in onnx format, some constant values are warpped with operation's attributes['value']
            # To move those constant value from numpy to device,
            # we have to move all the attributes['value'] of operation to device(if there is any).
            if operator.type == 'Constant':
                operator.attributes['value'] = convert_any_to_numpy(operator.attributes['value'])

        for _, variable in self._graph.variables.items():
            assert isinstance(variable, Variable), (
                f'Failed to send graph to device, incorrect variable {variable} found.'
            )
            variable.value = convert_any_to_numpy(variable.value, accept_none=True)

        return self

    def deploy(self, device: str):
        for _, operator in self._graph.operations.items():
            assert isinstance(operator, Operation), (
                f'Failed to send graph to device, incorrect operator {operator} found.'
            )

            # in onnx format, some constant values are warpped with operation's attributes['value']
            # To move those constant value from device to numpy,
            # we have to move all the attributes['value'] of operation to numpy(if there is any).
            if operator.type == 'Constant' and operator.platform != TargetPlatform.SOI:
                operator.attributes['value'] = convert_any_to_torch_tensor(
                    operator.attributes['value'], accept_none=False
                ).to(device)

            if operator.type == 'Constant' and operator.platform == TargetPlatform.SOI:
                value = operator.attributes['value']
                operator.attributes['value'] = convert_any_to_torch_tensor(value, accept_none=False, device='cpu')

            # PATCH 20220706, send quantization config to device.
            if isinstance(operator, QuantableOperation):
                for cfg, var in operator.config_with_variable:
                    if isinstance(cfg._scale, torch.Tensor):
                        cfg._scale = cfg._scale.to(device)
                    if isinstance(cfg._offset, torch.Tensor):
                        cfg._offset = cfg._offset.to(device)

        for _, variable in self._graph.variables.items():
            assert isinstance(variable, Variable), (
                f'Failed to send graph to device, incorrect variable {variable} found.'
            )
            # graph output variable has no destinations
            if len(variable.dest_ops) == 0:
                continue
            if variable.value is None:
                continue

            # check all destination operations platform are same.
            platforms = [op.platform for op in variable.dest_ops]
            if all([_ == platforms[0] for _ in platforms]) and platforms[0] == TargetPlatform.SOI:
                platform = TargetPlatform.SOI
            else:
                platform = TargetPlatform.UNSPECIFIED

            # if all downstream operations are shape related operations, send value to cpu
            if platform == TargetPlatform.SOI:
                variable.value = convert_any_to_torch_tensor(variable.value, accept_none=True).to('cpu')
            else:
                variable.value = convert_any_to_torch_tensor(variable.value, accept_none=True).to(device=device)

            # if variable is a shape-related variable, send it to cpu.
            if variable.is_parameter:
                if len(variable.dest_ops) > 1:
                    raise PermissionError(
                        f'PPQ can not process parameter variable({variable.name})'
                        f' with multiple destinations({[op.name for op in variable.dest_ops]}), split it first.'
                    )
                dest_op = variable.dest_ops[0]
                dest_idx = dest_op.inputs.index(variable)

                assert isinstance(dest_op, Operation)
                socket = dest_op.socket
                if socket.in_plat[dest_idx] == TargetPlatform.SOI:
                    variable.value = convert_any_to_torch_tensor(variable.value, accept_none=True).to('cpu')
        return self



import onnx
from onnx import helper, numpy_helper

from esp_ppq.core import NetworkFramework, QuantizationStates, convert_any_to_numpy
from esp_ppq.IR import BaseGraph, GraphExporter, Operation, QuantableVariable, Variable

MIN_SUFFIX = '_min'
MAX_SUFFIX = '_max'


class NxpExporter(GraphExporter):
    def __init__(self) -> None:
        super().__init__()

    def export_operation(self, operation: Operation) -> onnx.OperatorProto:
        op_proto = helper.make_node(
            op_type=operation.type,
            inputs=[_.name for _ in operation.inputs],
            outputs=[_.name for _ in operation.outputs],
            name=operation.name,
            **operation.attributes,
        )
        return op_proto

    def export_var(self, variable: Variable) -> onnx.TensorProto:
        if not variable.is_parameter:
            tensor_proto = helper.make_tensor_value_info(
                name=variable.name,
                # PPQ data type has exact same eunm value with onnx.
                elem_type=variable.dtype.value,
                shape=variable.shape,
            )
        else:
            value = convert_any_to_numpy(variable.value)
            if value is None:
                value = []
            else:
                value = value.flatten()
            tensor_proto = helper.make_tensor(
                name=variable.name, data_type=variable.dtype.value, dims=variable.shape, vals=value
            )
        return tensor_proto

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, export_param: bool = False):
        onnx_graph = onnx.GraphProto()
        onnx_graph.name = graph._name

        for operation in graph.operations.values():
            onnx_graph.node.append(self.export_operation(operation))

        for variable in graph.variables.values():
            tensor_proto = self.export_var(variable)

            if variable.name in graph.inputs:
                onnx_graph.input.append(tensor_proto)

            if variable.name in graph.outputs:
                onnx_graph.output.append(tensor_proto)

            if variable.is_parameter:
                onnx_graph.initializer.append(tensor_proto)

            if isinstance(variable, QuantableVariable):
                configs = variable.dest_op_configs + [variable.source_op_config]
                if variable.is_parameter and not export_param:
                    continue
                for config in configs:
                    if config is None:
                        continue  # source_op can be None
                    if config.can_export():
                        tensor_range = config.scale * pow(2, config.num_of_bits - 1)
                        min_val, max_val = -tensor_range, tensor_range - config.scale
                        min_tensor = numpy_helper.from_array(
                            convert_any_to_numpy(min_val).astype('float32'), variable.name + MIN_SUFFIX
                        )
                        min_info = helper.make_tensor_value_info(
                            variable.name + MIN_SUFFIX, min_tensor.data_type, min_tensor.dims
                        )

                        max_tensor = numpy_helper.from_array(
                            convert_any_to_numpy(max_val).astype('float32'), variable.name + MAX_SUFFIX
                        )
                        max_info = helper.make_tensor_value_info(
                            variable.name + MAX_SUFFIX, max_tensor.data_type, max_tensor.dims
                        )

                        onnx_graph.initializer.append(min_tensor)
                        onnx_graph.initializer.append(max_tensor)
                        onnx_graph.input.append(min_info)
                        onnx_graph.input.append(max_info)
                        break

        onnx_model = helper.make_model(onnx_graph)
        # onnx.checker.check_model(onnx_model)
        onnx.save(onnx_model, file_path)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class OptionalType(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = OptionalType()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsOptionalType(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # OptionalType
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OptionalType
    def ElemType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.TypeInfo import TypeInfo
            obj = TypeInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def OptionalTypeStart(builder):
    builder.StartObject(1)

def Start(builder):
    OptionalTypeStart(builder)

def OptionalTypeAddElemType(builder, elemType):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(elemType), 0)

def AddElemType(builder, elemType):
    OptionalTypeAddElemType(builder, elemType)

def OptionalTypeEnd(builder):
    return builder.EndObject()

def End(builder):
    return OptionalTypeEnd(builder)

import FlatBuffers.Dl.TypeInfo
try:
    from typing import Optional
except:
    pass

class OptionalTypeT(object):

    # OptionalTypeT
    def __init__(self):
        self.elemType = None  # type: Optional[FlatBuffers.Dl.TypeInfo.TypeInfoT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        optionalType = OptionalType()
        optionalType.Init(buf, pos)
        return cls.InitFromObj(optionalType)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, optionalType):
        x = OptionalTypeT()
        x._UnPack(optionalType)
        return x

    # OptionalTypeT
    def _UnPack(self, optionalType):
        if optionalType is None:
            return
        if optionalType.ElemType() is not None:
            self.elemType = FlatBuffers.Dl.TypeInfo.TypeInfoT.InitFromObj(optionalType.ElemType())

    # OptionalTypeT
    def Pack(self, builder):
        if self.elemType is not None:
            elemType = self.elemType.Pack(builder)
        OptionalTypeStart(builder)
        if self.elemType is not None:
            OptionalTypeAddElemType(builder, elemType)
        optionalType = OptionalTypeEnd(builder)
        return optionalType



from enum import Enum

from esp_ppq.core import OperationQuantizationConfig, TargetPlatform

from .graph import Operation, Variable


class GraphCommandType(Enum):
    # convert value inside graph to torch.tensor(usually from numpy)
    CONVERT_TO_TENSOR = 0

    # 图上权重部署到 GPU(tensor)，由 RunnableGraph 进行处理
    # deploy graph weights to GPU
    DEPLOY_TO_CUDA = 1

    # 图上权重部署到 CPU(tensor)，由 RunnableGraph 进行处理
    # deploy graph weights to CPU, in tensor format
    DEPLOY_TO_CPU = 2

    # 图上权重部署到 CPU(tensor)，由 RunnableGraph 进行处理
    # deploy graph weights to CPU, in array format
    DEPLOY_TO_NUMPY = 3

    # 图上权重部署到 MAC MPS(tensor)，由 RunnableGraph 进行处理
    # deploy graph weights to MAC MPS, in tensor format
    DEPLOY_TO_MPS = 4

    # 量化一个指定OP，同时将所有关联的 variable 转换为量化 variable
    # quantize a specified operation, and converts all connected
    # variables to quantable variables
    QUANTIZE_OPERATION = 5

    # 将一个OP的量化暂时解除，同时将所有关联的 variable 解除量化
    # deactivate quantization state of an op temporarily,
    # and deactivated all related variables
    DISABLE_OPERATION_QUANTIZATION = 6
    # 将一个OP的量化状态恢复
    # restore quantization state of a dequantized op
    RESTORE_OPERATION_QUANTIZATION = 7

    # 格式化 CLIP 算子，将不同行为的 CLIP 行为统一
    # regularize Clip operator
    FORMAT_CLIP = 9
    # 格式化 PAD 算子，将不同行为的 PAD 行为统一
    # regularize Pad operator
    FORMAT_PAD = 10
    # 格式化 GATHER 算子，将 index 参数（由input输入）植入算子属性
    # regularize gather operator
    FORMAT_GATHER = 11
    # 格式化 CAST 算子，统一 CAST 参数到 esp_ppq.core.DataType
    # regularize Cast operator
    FORMAT_CAST = 12
    # 格式化所有常量输入，尝试将他们转换为int32的
    # regularize all constant inputs
    FORMAT_INT64_CONSTANT = 13
    # 移除所有孤立节点
    # remove all isolated operators
    DELETE_ISOLATED = 14
    # 将所有参数变量进行分裂（只允许一个 dest_op ）
    # split variables and each variable is allowed to be used by only one operator
    FORMAT_PARAMETERS = 16

    # 用一个新的算子替换一个原来的
    # replace an old op with a new one
    REPLACE_OP = 17
    # 用一个新的 var 替换一个原来的
    # replace an old variable with a new one
    REPLACE_VAR = 18
    # 移除一个算子的输入参数，只能移除 Parameter
    # remove input parameters of am operator
    REMOVE_INPUT = 19

    # 图遍历模式匹配（要求返回路径）
    # graph traversal pattern matching(return paths)
    TRAVERSAL_PATTERN_MATCHING = 20
    # 图遍历模式匹配（仅要求返回点集）
    # graph traversal pattern matching(return op set)
    TRAVERSAL_OPSET_MATCHING = 21
    # 激活函数匹配
    # activation op matching
    ACTIVATION_MATCHING = 22
    # Concat 匹配
    # Concat op matching
    CONCAT_MATCHING = 23

    # 插入 Device Switcher
    # insert switcher
    INSERT_SWITCHER = 25
    # 移除 Device Switcher
    # remove switcher
    REMOVE_SWITCHER = 26

    # 融合图中的 计算层 与 BN
    # fuse Computing layer and BN
    FUSE_BN = 27
    # 删除图中的 Constant Input
    # remove constant input
    FORMAT_CONSTANT_INPUT = 28
    # 将 opset1 的 slice 弄成 opset 11 的
    FORMAT_SLICE = 29
    # 从一个指定位置将图截断
    TRUNCATE_ON_VAR = 30

    # 升级图中的 resize 到 opset 11
    FORMAT_RESIZE = 31

    # Replace Single Batchnorm to some op else.
    REPLACE_BATCHNORM_TO_CONV = 32
    REPLACE_BATCHNORM_TO_SCALE = 33

    # Fuse Bias add to Gemm, Conv and ConvTranspose
    FUSE_BIAS_ADD = 34

    # 移除 Identity
    # remove all identity ops from your graph
    REMOVE_IDENTITY = 35

    # Fuse splited swish
    FUSE_SWISH = 36


class GraphCommand:
    def __init__(self, command_type: GraphCommandType, **kwargs) -> None:
        assert isinstance(command_type, GraphCommandType), (
            f'Command Type must be a GraphCommandType object, but {type(command_type)} received.'
        )
        self.command_type = command_type
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f'GraphCommand object {self.__hash__()},\t Command type: {self.command_type},\t Args:{self.kwargs}'


class GraphDeployCommand(GraphCommand):
    def __init__(self, device: str) -> None:
        if device.startswith('cuda'):
            super().__init__(GraphCommandType.DEPLOY_TO_CUDA)
        elif device.startswith('cpu'):
            super().__init__(GraphCommandType.DEPLOY_TO_CPU)
        elif device.startswith('mps'):
            super().__init__(GraphCommandType.DEPLOY_TO_MPS)
        else:
            raise ValueError(f'Device type {device} not understand.')
        self._device = device

    def __str__(self) -> str:
        return super().__str__()


class QuantizeOperationCommand(GraphCommand):
    def __init__(self, op_name: str, target_platform: TargetPlatform, config: OperationQuantizationConfig) -> None:
        super().__init__(command_type=GraphCommandType.QUANTIZE_OPERATION)
        self.op_name = op_name
        self.target_platform = target_platform
        self.config = config


class ReplaceOperationCommand(GraphCommand):
    def __init__(self, op_name: str, replace_to: Operation) -> None:
        super().__init__(command_type=GraphCommandType.REPLACE_OP)
        self.op_name = op_name
        self.replace_to = replace_to


class ReplaceVariableCommand(GraphCommand):
    def __init__(self, var_name: str, replace_to: Variable) -> None:
        super().__init__(command_type=GraphCommandType.REPLACE_VAR)
        self.op_name = var_name
        self.replace_to = replace_to


class TruncateGraphCommand(GraphCommand):
    def __init__(self, var: Variable, mark_as_output: bool) -> None:
        super().__init__(command_type=GraphCommandType.TRUNCATE_ON_VAR)
        self.var = var
        self.mark_as_output = mark_as_output



import os
from typing import List

from esp_ppq.core import DataType, NetworkFramework, QuantizationProperty, QuantizationStates
from esp_ppq.IR import BaseGraph, GraphExporter, QuantableOperation

from .onnx_exporter import OnnxExporter
from .util import convert_value


class NCNNExporter(GraphExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        fd = open(config_path, 'w+')
        topo_order = graph.topological_sort()
        for op in topo_order:
            if op.is_computing_op and isinstance(op, QuantableOperation):
                fd.write(f'{op.name}_param_0 ')
                param_cfg = op.config.input_quantization_config[1]
                if not param_cfg.can_export():
                    continue

                assert (
                    param_cfg.state in {QuantizationStates.BAKED, QuantizationStates.ACTIVATED}
                    and param_cfg.observer_algorithm in {'minmax', 'Minmax'}
                    and param_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL)
                )
                # a workaround for depthwise conv in ncnn
                # will cause mis-alignment between ppq and ncnn
                if op.type == 'Conv' and op.attributes.get('group', 1) > 1:
                    group = op.attributes.get('group', 1)
                    scale = param_cfg.scale.reshape(group, -1).max(dim=1)[0]
                else:
                    scale = param_cfg.scale
                scale = convert_value(1 / scale, False, DataType.FP32)
                for s in scale:
                    fd.write('%f ' % s)
                fd.write('\n')

        for op in topo_order:
            if op.is_computing_op and isinstance(op, QuantableOperation):
                fd.write(f'{op.name} ')
                input_cfg = op.config.input_quantization_config[0]
                assert input_cfg.state == QuantizationStates.ACTIVATED and input_cfg.policy.has_property(
                    QuantizationProperty.PER_TENSOR
                )
                scale = convert_value(1 / input_cfg.scale, True, DataType.FP32)
                fd.write('%f ' % scale)
                fd.write('\n')
        fd.close()

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        input_shapes: List[List[int]] = [[1, 3, 224, 224]],
    ):
        if config_path is not None:
            self.export_quantization_config(config_path, graph)

        _, ext = os.path.splitext(file_path)
        if ext == '.onnx' or graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



import torch
import torchvision
from trainer import ImageNetTrainer
from Utilities.Imagenet import *  # check esp_ppq.samples.imagenet.Utilities
from Utilities.Imagenet.imagenet_util import load_imagenet_from_directory  # check esp_ppq.samples.imagenet.Utilities

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_native_graph, load_torch_model
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.quantization.optim import *
from esp_ppq.quantization.quantizer import TensorRTQuantizer

"""
    使用这个脚本来尝试在 Imagenet 数据集上执行量化感知训练
        使用 imagenet 中的数据测试量化精度与 calibration
        默认的 imagenet 数据集位置:Assets/Imagenet_Train, Assets/Imagenet_Valid
        你可以通过软连接创建它们:
            ln -s /home/data/Imagenet/val Assets/Imagenet_Valid
            ln -s /home/data/Imagenet/train Assets/Imagenet_Train
"""

CFG_DEVICE = 'cuda'  # 一个神奇的字符串，用来确定执行设备
CFG_BATCHSIZE = 32  # 测试与calib时的 batchsize
CFG_INPUT_SHAPE = (CFG_BATCHSIZE, 3, 224, 224)  # 用来确定模型输入的尺寸，好像 imagenet 都是这个尺寸
CFG_VALIDATION_DIR = 'Assets/Imagenet_Valid'  # 用来读取 validation dataset
CFG_TRAIN_DIR = 'Assets/Imagenet_Train'  # 用来读取 train dataset，注意该集合将被用来 calibrate 你的模型
CFG_PLATFORM = TargetPlatform.TRT_INT8  # 用来指定目标平台

# ------------------------------------------------------------
# 在这个例子中我们将向你展示如何在 PPQ 中对你的网络进行量化感知训练
# 你可以使用带标签的数据执行正常的训练流程，也可以使用类似蒸馏的方式进行无标签训练
# PPQ 模型的训练过程与 Pytorch 遵循相同的逻辑，你可以使用 Pytorch 中的技巧来获得更好的训练效果
# ------------------------------------------------------------
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
graph = load_torch_model(model=model, sample=torch.zeros([1, 3, 224, 224]).cuda())

# ------------------------------------------------------------
# 我们首先进行标准的量化流程，为所有算子初始化量化信息，并进行 Calibration
# ------------------------------------------------------------
quantizer = TensorRTQuantizer(graph=graph)
dispatching_table = PFL.Dispatcher(graph=graph).dispatch(quantizer.quant_operation_types)

# 为算子初始化量化信息
for op in graph.operations.values():
    quantizer.quantize_operation(op_name=op.name, platform=dispatching_table[op.name])

executor = TorchExecutor(graph=graph)
executor.tracing_operation_meta(inputs=torch.zeros([1, 3, 224, 224]).cuda())

calib_dataloader = load_imagenet_from_directory(
    directory=CFG_TRAIN_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, subset=1280, require_label=True, num_of_workers=8
)

training_dataloader = load_imagenet_from_directory(
    directory=CFG_TRAIN_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, require_label=True, num_of_workers=8
)

eval_dataloader = load_imagenet_from_directory(
    directory=CFG_VALIDATION_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, require_label=True, num_of_workers=8
)

with ENABLE_CUDA_KERNEL():
    # ------------------------------------------------------------
    # 创建优化管线，由于后续还要继续训练我们的模型，我们不能在此处调用
    # ParameterBakingPass()，一旦模型权重完成烘焙，则它们不能被进一步调整
    # ------------------------------------------------------------
    pipeline = PFL.Pipeline(
        [
            QuantizeSimplifyPass(),
            QuantizeFusionPass(activation_type=quantizer.activation_fusion_types),
            ParameterQuantizePass(),
            RuntimeCalibrationPass(method='kl'),
            PassiveParameterQuantizePass(),
            QuantAlignmentPass(),
            # LearnedStepSizePass(steps=500, block_size=5)
        ]
    )
    pipeline.optimize(
        calib_steps=8, collate_fn=lambda x: x[0].cuda(), graph=graph, dataloader=calib_dataloader, executor=executor
    )

    # ------------------------------------------------------------
    # 完成量化后，我们将开始进行 QAT 的模型训练，我们希望你能够注意到：
    #
    # 1. 不能从零开始使用 QAT 的方法完成训练，你应当先训练好浮点的模型，或者在一个预训练的模型基础上进行 QAT finetuning.
    # 2. 你必须完成标准量化流程
    # 3. PPQ Executor 长得很像 Pytorch Module，单机训练应该不会遇到太多困难，但它不支持多卡训练

    # 训练的代码我们封装进了一个叫做 ImageNetTrainer 的东西
    # 你可以打开它看到具体的训练逻辑
    # ------------------------------------------------------------
    trainer = ImageNetTrainer(graph=graph)

    best_acc = 0
    for epoch in range(20):
        trainer.epoch(training_dataloader)
        current_acc = trainer.eval(eval_dataloader)
        if current_acc > best_acc:
            trainer.save('Best.native')

    graph = load_native_graph(import_file='Best.native')
    PFL.Exporter(platform=CFG_PLATFORM).export(file_path='export.onnx', graph=graph, config_path='export.json')



# https://onnxruntime.ai/docs/api/python/auto_examples/plot_profiling.html
from time import time

import numpy as np
import onnxruntime as ort
from tqdm import tqdm


def Benchmark(onnx_file: str, steps: int = 10000, providers=['CUDAExecutionProvider'], provider_options=None) -> float:
    """Benckmark with Onnxruntime

    * Quantized Model of Onnxruntime - TensorrtExecutionProvider and Onnxruntime - CUDAExecutionProvider has different format.

    * Onnx that generated by PPQ is not supportable with TensorrtExecutionProvider.

    * Set providers=CUDAExecutionProvider before benchmark this file.

    """
    sess = ort.InferenceSession(path_or_bytes=onnx_file, providers=providers, provider_options=provider_options)

    feed_dict, output_names = {}, []
    for input_meta in sess.get_inputs():
        name, dtype, shape = input_meta.name, input_meta.type, input_meta.shape

        for element in shape:
            if element is None or type(element) == str:
                raise TypeError('Dynamic input is not supported by this function.')

        if dtype == 'tensor(float)':
            feed_dict[name] = np.random.random(size=shape).astype(np.float32)
        else:
            raise Exception(f'Input {name} has unexpected data type.')

    for output_meta in sess.get_outputs():
        output_names.append(output_meta.name)

    tick = time()
    for _ in tqdm(range(steps)):
        sess.run(output_names=output_names, input_feed=feed_dict)
    tok = time()

    print(f'Time span: {tok - tick: .4f} sec')
    return tick - tok


def Profile(onnx_file: str, steps: int = 1, providers=['CUDAExecutionProvider'], provider_options=None):
    """Profile with Onnxruntime

    * Quantized Model of Onnxruntime - TensorrtExecutionProvider and Onnxruntime - CUDAExecutionProvider has different format.

    * Onnx that generated by PPQ is not supportable with TensorrtExecutionProvider.

    * Set providers=CUDAExecutionProvider before benchmark this file.

    """
    options = ort.SessionOptions()
    options.enable_profiling = True
    sess = ort.InferenceSession(
        path_or_bytes=onnx_file, providers=providers, provider_options=provider_options, sess_options=options
    )

    feed_dict, output_names = {}, []
    for input_meta in sess.get_inputs():
        name, dtype, shape = input_meta.name, input_meta.type, input_meta.shape

        for element in shape:
            if element is None or type(element) == str:
                raise TypeError('Dynamic input is not supported by this function.')

        if dtype == 'tensor(float)':
            feed_dict[name] = np.random.random(size=shape).astype(np.float32)
        else:
            raise Exception(f'Input {name} has unexpected data type.')

    for output_meta in sess.get_outputs():
        output_names.append(output_meta.name)

    for _ in tqdm(range(steps)):
        sess.run(output_names=output_names, input_feed=feed_dict)

    prof_file = sess.end_profiling()
    print(f'Profile file is generated at {prof_file}, open it with your web browser chrome://tracing/')



from esp_ppq.lib import (
    register_calibration_observer,
    register_network_exporter,
    register_network_parser,
    register_network_quantizer,
    register_operation_handler,
)

from .espdl_interface import (
    espdl_quantize_onnx,
    espdl_quantize_torch,
    generate_test_value,
    get_target_platform,
    insert_streaming_cache_on_var,
)
from .fsys import (
    compare_cosine_similarity_between_results,
    create_dir,
    dump_internal_results,
    dump_to_file,
    load_calibration_dataset,
    load_from_file,
    split_result_to_directory,
)
from .interface import (
    DISABLE_CUDA_KERNEL,
    ENABLE_CUDA_KERNEL,
    UnbelievableUserFriendlyQuantizationSetting,
    dispatch_graph,
    dump_torch_to_onnx,
    empty_ppq_cache,
    export,
    export_ppq_graph,
    format_graph,
    load_graph,
    load_native_graph,
    load_onnx_graph,
    load_torch_model,
    manop,
    quantize,
    quantize_native_model,
    quantize_onnx_model,
    quantize_torch_model,
)
from .setting import (
    ActivationQuantizationSetting,
    BiasCorrectionSetting,
    BlockwiseReconstructionSetting,
    ChannelSplitSetting,
    DispatchingTable,
    EqualizationSetting,
    GraphFormatSetting,
    LSQSetting,
    ParameterQuantizationSetting,
    QuantizationFusionSetting,
    QuantizationSetting,
    QuantizationSettingFactory,
    SSDEqualizationSetting,
    TemplateSetting,
    WeightSplitSetting,
)



# Test Quantization System Performace on MMPose Models with Coco 2017 Dataset
#
#   1. How to use:
#      Run this script with python directly.
#

# Quantizer Configuration
SYMMETRICAL = True
POWER_OF_2 = False
PERCHANNEL = True
BIT_WIDTH = 8
FP8 = False

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/mmpose'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/Pose/Calib'
TEST_DIR = 'QuantZoo/Data/Pose/Test'

# calibration & test batchsize
BATCHSIZE = 1

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {'Model': 'deeppose_res50_coco_256x192', 'Output': ['/backbone/layer4/layer4.2/relu_2/Relu_output_0']},
    {'Model': 'hrnet_w32', 'Output': ['onnx::Conv_2946']},
    {'Model': 'simcc-mobilenet_v2', 'Output': ['537']},
    {'Model': 'rtmpose-s', 'Output': ['onnx::Conv_439']},
]

import os

import torch
from QuantZoo.Data.Pose.Data import load_coco_keypoint_dataset
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze, report

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import BaseGraph, SearchableGraph
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)

calib_dataloader = load_coco_keypoint_dataset(data_dir=CALIB_DIR)
test_dataloader = load_coco_keypoint_dataset(data_dir=TEST_DIR)


def output_ops(graph: BaseGraph, output_vars: list):
    """
    Return all ops that under the given output variable,
    We perfer to dequantize those ops.
    """
    serach_engine = SearchableGraph(graph)
    source_ops = [graph.variables[var].source_op for var in output_vars]
    return serach_engine.opset_matching(
        sp_expr=lambda x: x in source_ops,
        rp_expr=lambda x, y: True,
        ep_expr=None,
        direction='down',
    )


with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))
        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH, per_channel=PERCHANNEL
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph)

        # convert op to quantable-op
        non_quantable_ops = output_ops(graph=graph, output_vars=monitoring_vars)
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm'} and op not in non_quantable_ops:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cpu')
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 256, 192]).cuda())

        pipeline.optimize(
            graph=graph,
            dataloader=calib_dataloader,
            verbose=True,
            calib_steps=32,
            collate_fn=lambda x: x[0].to('cuda'),
            executor=executor,
        )

        # error analyze
        performance = error_analyze(
            graph=graph,
            outputs=monitoring_vars,
            dataloader=test_dataloader,
            collate_fn=lambda x: x[0].cuda(),
            verbose=True,
        )



import torch

from esp_ppq.core import QuantizationProperty, TensorQuantizationConfig, convert_any_to_torch_tensor
from esp_ppq.IR import BaseGraph, Operation, Variable
from esp_ppq.utils.round import ppq_tensor_round

from .onnxruntime_exporter import ONNXRUNTIMExporter


class OpenvinoExporter(ONNXRUNTIMExporter):
    """Openvino 喜欢所有 QuantizeLinear 上面都带 axis 属性，所以这里需要单独处理一下"""

    def insert_quantize_node(
        self, graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of Quantize Node: QuantizeLinear and QuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> INT
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type='QuantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = 0

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].dtype = value_type
            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> FP8
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='QuantizeFloating',
                attributes={
                    'min': config.quant_min,
                    'max': config.quant_max,
                    'exponent': config.exponent_bits,
                    'mantissa': config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        else:
            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.'
            )

    def insert_dequantize_node(
        self, graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of DeQuantize Node: DeQuantizeLinear and DeQuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type='DequantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = 0

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.inputs[0].dtype = value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = torch.float32

            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='DequantizeFloating',
                attributes={
                    'min': config.quant_min,
                    'max': config.quant_max,
                    'exponent': config.exponent_bits,
                    'mantissa': config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else:
                created.attributes['axis'] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape

            return created

        else:
            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.'
            )



import torch

from esp_ppq.core import (
    OBSERVER_FLOATING_MSE_FETCHES,
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
)
from esp_ppq.IR import Variable
from esp_ppq.quantization.qfunction import PPQuantFunction
from esp_ppq.utils.fetch import channel_random_fetch, tensor_random_fetch

from .base import BaseTensorObserver


class ConstantObserver(BaseTensorObserver):
    """
    This observer will directly return scale = 1, offset = 0
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        self._value_shape = None
        self._value_device = None

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        self._value_shape = value.shape  # Do nothing here.
        self._value_device = value.device

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        device = self._value_device
        if self._quant_cfg.policy.has_property(QuantizationProperty.FLOATING):
            if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                self._quant_cfg.scale = torch.tensor([1.0], dtype=torch.float32, device=device).squeeze(0)
                self._quant_cfg.offset = torch.tensor([0.0], dtype=torch.float32, device=device).squeeze(0)
                self._quant_cfg.state = QuantizationStates.ACTIVATED

            elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                num_of_elements = self._value_shape[self._quant_cfg.channel_axis]
                scales = [1.0 for _ in range(num_of_elements)]
                offsets = [0.0 for _ in range(num_of_elements)]

                self._quant_cfg.scale = torch.tensor(scales, dtype=torch.float32, device=device)
                self._quant_cfg.offset = torch.tensor(offsets, dtype=torch.float32, device=device)
                self._quant_cfg.state = QuantizationStates.ACTIVATED
        else:
            raise TypeError('This Observer is designed for floating quantization.')


class DirectMSEObserver(BaseTensorObserver):
    """
    Direct MSE Observer, this observer compute MSE Loss directly.
    In PPQ there is another version of MSE Observer called TorchMSEObserver,
        which uses histogram for accelerating the computation of MSE Loss.

    We prefer to implements a direct MSE Observer for floating quantization,
        cause floating quantization can have only a few candidates of scale, which
        greatly reduce the computation complexity of solving MSE Loss.
    """

    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)

        if not quant_cfg.policy.has_property(QuantizationProperty.FLOATING):
            raise TypeError('MSE Floating Observer is designed for floating quantization.')

        if not quant_cfg.policy.has_property(QuantizationProperty.POWER_OF_2):
            raise TypeError('MSE Floating Observer is designed for power-of-2 quantization.')

        self._value_shape = None
        self._value_device = None
        self._collector = []
        self._fetches = OBSERVER_FLOATING_MSE_FETCHES

    @torch.no_grad()
    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            if self._watch_on.is_parameter:
                value = torch.transpose(value, dim0=0, dim1=self._quant_cfg.channel_axis)
                self._collector.append(torch.flatten(value, start_dim=1))

            else:
                self._collector.append(channel_random_fetch(value, fetchs_per_channel=self._fetches))
                self._value_device = value.device

        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            self._collector.append(tensor_random_fetch(value, num_of_fetches=self._fetches))

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        device = self._value_device
        scale_candidates = [0.0078125, 0.03125, 0.125, 1.0, 4.0, 16.0, 64.0]

        if not self._collector:
            raise PermissionError(
                'Observer collector is empty, you should invoke observe function before render quantization config.'
            )

        self._quant_cfg.state = QuantizationStates.ACTIVATED
        if self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
            self._collector = torch.cat(self._collector, dim=0)
            num_of_channel = self._collector.shape[0]
            scales = torch.ones(size=[num_of_channel], dtype=torch.float32, device=device)
            offsets = torch.zeros(size=[num_of_channel], dtype=torch.float32, device=device)

            losses = []
            for scale in scale_candidates:
                self._quant_cfg._scale = scales * scale
                self._quant_cfg._offset = offsets

                qt = PPQuantFunction(self._collector, self._quant_cfg)
                fp = self._collector
                loss = torch.mean(torch.square(qt - fp), dim=-1, keepdim=True)
                losses.append(loss)

            scale_index = torch.argmin(torch.cat(losses, dim=-1), dim=-1)
            best_scales = []
            for index in scale_index.cpu():
                best_scales.append(scale_candidates[index])
            self._quant_cfg._scale = torch.tensor(best_scales).to(self._collector.device)

        elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            self._collector = torch.cat(self._collector, dim=0)
            scales = torch.ones(size=[1], dtype=torch.float32, device=device)
            offsets = torch.zeros(size=[1], dtype=torch.float32, device=device)

            losses = []
            for scale in scale_candidates:
                self._quant_cfg._scale = scales * scale
                self._quant_cfg._offset = offsets

                qt = PPQuantFunction(self._collector, self._quant_cfg)
                fp = self._collector
                loss = torch.mean(torch.square(qt - fp))

                losses.append((loss.item(), scale))

            best_scale = sorted(losses)[0][1]
            self._quant_cfg._scale = scales * best_scale



import argparse
import json
import os

import qti.aisw.dlc_utils as dlc
import snpe

parser = argparse.ArgumentParser(description='Write ppq qparams to snpe dlc')
parser.add_argument('--input_dlc_model', default='snpe_quantized.dlc', help='path to snpe quantized dlc model')
parser.add_argument('--output_dlc_model', default='ppq_export.dlc', help='path to export quantized dlc')
parser.add_argument('--qparam', default='quantized.json', help='path to ppq qparams json')


def json_load(filename):
    with open(filename) as json_file:
        data = json.load(json_file)
    return data


def write_qparams_to_dlc_model(input_dlc, output_dlc, activation_qparams):
    model = dlc.modeltools.Model()
    model.load(input_dlc)
    model.set_tf_encoding_type("TF")

    for snpe_layer in model.get_layers():
        print('\n write qparams to {}'.format(snpe_layer['name']))
        for snpe_layer_out_ind, snpe_layer_out in enumerate(snpe_layer['output_names']):
            layer_name = snpe_layer['name']
            print(
                'original quant encodings : ',
                model.get_tf_output_encoding_by_index(name=layer_name, index=snpe_layer_out_ind),
            )
            top = snpe_layer['output_names'][0]

            if top not in activation_qparams.keys():
                # Before the Reshape layer, SNPE will insert the shape conversion layer(xxx.ncs)
                # Because the SNPE data is arranged as NHWC
                assert top.endswith('.ncs'), '{} ranges not exists'.format(top)
                bottom = snpe_layer['input_names'][0]
                new_enc = activation_qparams[bottom][0]  # List[dict]
            else:
                new_enc = activation_qparams[top][0]  # List[dict]

            model.set_tf_output_encoding_by_index(
                name=layer_name, index=snpe_layer_out_ind, bitwidth=8, min=new_enc["min"], max=new_enc["max"]
            )
            print(
                'ppq quant encodings : ',
                model.get_tf_output_encoding_by_index(name=layer_name, index=snpe_layer_out_ind),
            )
    model.quantize_weights(should_quantize=True)
    model.save(output_dlc)


if __name__ == '__main__':
    args = parser.parse_args()
    act_ranges = json_load(args.qparam)['activation_encodings']
    write_qparams_to_dlc_model(args.input_dlc_model, args.output_dlc_model, act_ranges)
